from typing import List, Optional

from acapelladb.Entry import Entry
from acapelladb.utils.http import AsyncSession, raise_if_error


class Transaction(object):
    def __init__(self, session: AsyncSession, index: int, api_prefix: str):
        """        
        Создание транзакции. Этот метод предназначен для внутреннего использования.
        """
        self._session = session
        self._index = index
        self._completed = False
        self._api_prefix = api_prefix

    async def commit(self):
        """
        Применение транзакции. Применить/откатить транзакцию можно только один раз.
        
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, не найдена 
        :raise TransactionCompletedError: когда транзакция, уже завершена
        :raise KvError: когда произошла неизвестная ошибка на сервере         
        """
        if not self._completed:
            response = await self._session.post(f'{self._api_prefix}/v2/tx/{self._index}/commit')
            raise_if_error(response.status)
            self._completed = True

    async def rollback(self):
        """
        Откат транзакции. Применить/откатить транзакцию можно только один раз.
        Если транзакция уже завершена, откат не выполняется, но не бросает искличение.
        
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        if not self._completed:
            response = await self._session.post(f'{self._api_prefix}/v2/tx/{self._index}/rollback')
            raise_if_error(response.status)
            self._completed = True

    async def keep_alive(self):
        """
        Продление жизни транзакции.
        Этот запрос необходим, чтобы определять зависшие транзакции.
        
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, не найдена 
        :raise TransactionCompletedError: когда транзакция, уже завершена
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        response = await self._session.post(f'{self._api_prefix}/v2/tx/{self._index}/keep-alive')
        raise_if_error(response.status)

    async def get_entry(self, partition: List[str], clustering: Optional[List[str]] = None,
                        watch: bool = False, n: int = 3, r: int = 2, w: int = 2) -> Entry:
        """        
        Получение значения по указанному ключу в транзакции.
        
        :param partition: распределительный ключ
        :param clustering: сортируемый ключ
        :param watch: если true, то блокирует ключ до конца транзакции
        :param n: количество реплик
        :param r: количество ответов для подтверждения чтения
        :param w: количество ответов для подтверждения записи
        :return: Entry для указанного ключа с полученным значением
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, в которой выполняется операция, не найдена 
        :raise TransactionCompletedError: когда транзакция, в которой выполняется операция, уже завершена
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        entry = self.entry(partition, clustering, n, r, w)
        await entry.get(watch)
        return entry

    def entry(self, partition: List[str], clustering: Optional[List[str]] = None,
              n: int = 3, r: int = 2, w: int = 2) -> Entry:
        """
        Создание Entry для указанного ключа в транзакции. Не выполняет никаких запросов.
        Можно использовать, если нет необходимости знать текущие значение и версию.
        
        :param partition: распределительный ключ
        :param clustering: сортируемый ключ
        :param n: количество реплик
        :param r: количество ответов для подтверждения чтения
        :param w: количество ответов для подтверждения записи
        :return: Entry для указанного ключа
        """
        clustering = clustering or []
        return Entry(self._session, self._api_prefix, partition, clustering, 0, None, n, r, w, self._index)

    @property
    def index(self) -> int:
        """
        :return: индекс транзакции 
        """
        return self._index



from typing import List, Optional

from acapelladb.utils.collections import remove_none_values
from acapelladb.utils.http import AsyncSession, key_to_str, raise_if_error
from acapelladb.utils.assertion import check_key, check_nrw


class Cursor(object):
    def __init__(self, session: AsyncSession, tree: List[str], key: List[str], version: int, value: Optional[any],
                 node: Optional[str], n: int, r: int, w: int, transaction: Optional[int], api_prefix: str):
        check_key(tree)
        check_key(key)
        check_nrw(n, r, w)
        self._tree = tree
        self._session = session
        self._key = key
        self._version = version
        self._value = value
        self._node = node
        self._n = n
        self._r = r
        self._w = w
        self._transaction = transaction
        self._api_prefix = api_prefix

    async def get(self) -> Optional[any]:
        """
        Запрашивает текущее значение с сервера.
        Запоминает новые значение и версию.

        :return: полученное значение
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, в которой выполняется операция, не найдена 
        :raise TransactionCompletedError: когда транзакция, в которой выполняется операция, уже завершена
        :raise TreeNotFoundError: когда не найдено дерево
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        response = await self._session.get(
            f'{self._api_prefix}/v2/dt/{key_to_str(self._tree)}/keys/{key_to_str(self._key)}',
            params=remove_none_values({
                'n': self._n,
                'r': self._r,
                'w': self._w,
                'transaction': self._transaction,
            })
        )
        raise_if_error(response.status)

        body = await response.json()
        # TODO: версия в DT
        # self._version = int(body['version'])
        self._value = body.get('value')
        return self._value

    async def set(self, new_value: Optional[any]) -> int:
        """
        Устанавливает новое значение.
        Запоминает новые значение и версию.

        :param new_value: новое значение
        :return: новая версия
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, в которой выполняется операция, не найдена 
        :raise TransactionCompletedError: когда транзакция, в которой выполняется операция, уже завершена
        :raise TreeNotFoundError: когда не найдено дерево
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        response = await self._session.put(
            f'{self._api_prefix}/v2/dt/{key_to_str(self._tree)}/keys/{key_to_str(self._key)}',
            params=remove_none_values({
                'n': self._n,
                'r': self._r,
                'w': self._w,
                'transaction': self._transaction,
            }),
            json=new_value
        )
        raise_if_error(response.status)

        # TODO: версия в DT
        self._value = new_value
        # body = response.json()
        # self._version = int(body['version'])
        return self._version

    async def next(self) -> Optional['Cursor']:
        """
        Поиск следующего за текущим ключа.
        :return: Cursor со следующим ключом, None - если ключ не найден.
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, в которой выполняется операция, не найдена 
        :raise TransactionCompletedError: когда транзакция, в которой выполняется операция, уже завершена
        :raise TreeNotFoundError: когда не найдено дерево
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        response = await self._session.get(
            f'{self._api_prefix}/v2/dt/{key_to_str(self._tree)}/keys/{key_to_str(self._key)}/next',
            params=remove_none_values({
                'n': self._n,
                'r': self._r,
                'w': self._w,
                'transaction': self._transaction,
                'node': self._node,
            })
        )
        if response.status == 404:
            return None
        raise_if_error(response.status)

        body = await response.json()
        return Cursor(self._session, self._tree, body['key'], 0, body.get('value'), body.get('node'),
                      self._n, self._r, self._w, self._transaction, self._api_prefix)

    async def prev(self) -> Optional['Cursor']:
        """
        Поиск предыдущего перед текущим ключа.
        :return: Cursor с предыдущим ключом, None - если ключ не найден.
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, в которой выполняется операция, не найдена 
        :raise TransactionCompletedError: когда транзакция, в которой выполняется операция, уже завершена
        :raise TreeNotFoundError: когда не найдено дерево
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        response = await self._session.get(
            f'{self._api_prefix}/v2/dt/{key_to_str(self._tree)}/keys/{key_to_str(self._key)}/prev',
            params=remove_none_values({
                'n': self._n,
                'r': self._r,
                'w': self._w,
                'transaction': self._transaction,
                'node': self._node,
            })
        )
        if response.status == 404:
            return None
        raise_if_error(response.status)

        body = await response.json()
        return Cursor(self._session, self._tree, body['key'], 0, body.get('value'), body.get('node'),
                      self._n, self._r, self._w, self._transaction, self._api_prefix)

    @property
    def value(self) -> Optional[any]:
        """
        :return: значение 
        """
        return self._value

    @property
    def version(self) -> int:
        """
        :return: версия 
        """
        return self._version

    @property
    def key(self) -> List[str]:
        """
        :return: ключ 
        """
        return self._key



from typing import List, Optional, Dict

from acapelladb import Entry
from acapelladb.IndexField import IndexField
from acapelladb.utils.http import AsyncSession, raise_if_error, key_to_str


class QueryCondition(object):
    def __init__(self, eq: Optional[any] = None, from_: Optional[any] = None, to_: Optional[any] = None):
        self.eq = eq
        self.from_ = from_
        self.to_ = to_

    def __repr__(self) -> str:
        return f'QueryCondition(eq={self.eq}, from={self.from_}, to={self.to_}'

    def to_json(self):
        return {
            'eq': self.eq,
            'from': self.from_,
            'to': self.to_
        }


class PartitionIndex(object):
    def __init__(self, session: AsyncSession, partition: List[str], api_prefix: str):
        assert len(partition) >= 2, "Indexed partition must be in format: [<user>, <keyspace>, ...]"

        self._session = session
        self._user = partition[0]
        self._keyspace = partition[1]
        self._partition = partition
        self._api_prefix = api_prefix

    async def query(self, query: Dict[str, QueryCondition], limit: Optional[int] = None) -> List[Entry]:
        url = f'{self._api_prefix}/v2/kv/partition/{key_to_str(self._partition)}/index-query'
        response = await self._session.get(url, json={
            'params': {
               'limit': limit
            },
            'query': {field: cond.to_json() for field, cond in query.items()}
        })
        raise_if_error(response.status)
        data = await response.json()
        return [Entry(self._session, self._api_prefix, self._partition, e['key'], 0, e.get('value'), 3, 2, 2, None) for e in data]

    async def set_index(self, tag: int, fields: List[IndexField]):
        url = f'{self._api_prefix}/v2/users/{self._user}/keyspaces/{self._keyspace}/indexes/{tag}'
        response = await self._session.put(url, json={
            'fields': [f.to_json() for f in fields]
        })
        raise_if_error(response.status)

    async def get_indexes(self) -> Dict[int, List[IndexField]]:
        url = f'{self._api_prefix}/v2/users/{self._user}/keyspaces/{self._keyspace}/indexes'
        response = await self._session.get(url)
        raise_if_error(response.status)
        data = await response.json()
        indexes = data['indexes']
        return {int(tag): [IndexField.from_json(field) for field in index['fields']] for tag, index in indexes.items()}



class KvError(Exception):
    pass


class CasError(KvError):
    pass


class KeyNotFoundError(KvError):
    pass


class TreeNotFoundError(KvError):
    pass


class TransactionNotFoundError(KvError):
    pass


class TransactionCompletedError(KvError):
    pass


class AuthenticationFailedError(KvError):
    pass



from typing import Optional, List


class BatchBase(object):
    """
    Базовый класс для формирования батч-запрсов.
    """

    async def set(self, partition: List[str], clustering: List[str], new_value: Optional[any],
                  reindex: bool, n: int, r: int, w: int) -> int:
        """
        Добавляет set-запрос в батч. Метод возвращает управление, когда батч будет выполнен.
        :param partition: распределительный ключ
        :param clustering: сортируемый ключ
        :param new_value: новое значение
        :param reindex: переиндексировать ключ с новым значением?
        :param n: количество реплик
        :param r: количество ответов для подтверждения чтения
        :param w: количество ответов для подтверждения записи
        :return: новая версия
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        raise NotImplementedError()

    async def cas(self, partition: List[str], clustering: List[str], new_value: Optional[any],
                  old_version: int, reindex: bool, n: int, r: int, w: int) -> int:
        """
        Добавляет cas-запрос в батч. Метод возвращает управление, когда батч будет выполнен.
        :param partition: распределительный ключ
        :param clustering: сортируемый ключ
        :param new_value: новое значение
        :param old_version: версия для сравнения
        :param reindex: переиндексировать ключ с новым значением?
        :param n: количество реплик
        :param r: количество ответов для подтверждения чтения
        :param w: количество ответов для подтверждения записи
        :return: новая версия
        :raise CasError: когда текущая версия значения не совпала с указанной
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        raise NotImplementedError()



from asyncio import Future
from typing import List, Optional, Dict, Tuple

import asyncio

from acapelladb.BatchBase import BatchBase
from acapelladb.utils.http import AsyncSession, key_to_str, raise_if_error


class BatchEntry(object):
    def __init__(self, new_value: Optional[any], old_version: Optional[int], reindex: bool):
        self.new_value = new_value
        self.old_version = old_version
        self.reindex = reindex
        self.new_version = 0  # for response


class PartitionBatch(object):
    def __init__(self, n: int, r: int, w: int):
        self.n = n
        self.r = r
        self.w = w
        self.batch: Dict[Tuple[str, ...], BatchEntry] = {}

    def set(self, clustering: List[str], new_value: Optional[any], reindex: bool) -> BatchEntry:
        key = tuple(clustering)
        entry = BatchEntry(new_value, None, reindex)
        self.__assert_not_set(key)
        self.batch[key] = entry
        return entry

    def cas(self, clustering: List[str], new_value: Optional[any], old_version: int, reindex: bool) -> BatchEntry:
        key = tuple(clustering)
        entry = BatchEntry(new_value, old_version, reindex)
        self.__assert_not_set(key)
        self.batch[key] = entry
        return entry

    def need_reindex(self) -> bool:
        return any([e.reindex for e in self.batch.values()])

    def build_request_body(self) -> object:
        return [
            {
                'key': k,
                'value': e.new_value,
                'version': e.old_version
            }
            for k, e in self.batch.items()
        ]

    def apply_response(self, body: List[dict]):
        assert len(body) == len(self.batch)

        for item in body:
            key = tuple(item['key'])
            new_version = item['version']
            self.batch[key].new_version = new_version

    def __assert_not_set(self, clustering: Tuple[str, ...]):
        assert clustering not in self.batch, "Key can be added to batch only one time"


class BatchManual(BatchBase):
    def __init__(self, session: AsyncSession, api_prefix: str):
        self._session = session
        self._future = Future()
        self._in_process = True
        self._batch: Dict[Tuple[str, ...], PartitionBatch] = {}
        self._api_prefix = api_prefix

    async def set(self, partition: List[str], clustering: List[str], new_value: Optional[any],
                  reindex: bool, n: int, r: int, w: int) -> int:
        self._assert_in_process()
        partition_batch = self._batch.setdefault(tuple(partition), PartitionBatch(n, r, w))
        entry = partition_batch.set(clustering, new_value, reindex)
        await self._future
        return entry.new_version

    async def cas(self, partition: List[str], clustering: List[str], new_value: Optional[any], old_version: int,
                  reindex: bool, n: int, r: int, w: int) -> int:
        self._assert_in_process()
        partition_batch = self._batch.setdefault(tuple(partition), PartitionBatch(n, r, w))
        entry = partition_batch.cas(clustering, new_value, old_version, reindex)
        await self._future
        return entry.new_version

    async def send(self):
        # если в батче ничего нет, то скорее всего все запросы были
        # закинуты в event-loop и ещё не выполнились, так что нужно
        # дать им возможность это сделать
        if len(self._batch) == 0:
            # более-менее стандартный способ сделать yield
            await asyncio.sleep(0)

        self._in_process = False
        requests = []

        for k, b in self._batch.items():
            requests.append(self._send_partition(k, b))
        await asyncio.gather(*requests)

        self._future.set_result(None)

    async def _send_partition(self, partition: Tuple[str, ...], batch: PartitionBatch):
        url = f'{self._api_prefix}/v2/kv/partition/{key_to_str(partition)}'
        response = await self._session.put(url, params={
            'n': batch.n,
            'r': batch.r,
            'w': batch.w,
            'reindex': str(batch.need_reindex())
        }, json=batch.build_request_body())
        raise_if_error(response.status)
        batch.apply_response(await response.json())

    def _assert_in_process(self):
        assert self._in_process, "Batch can be used only one time"



import asyncio
from asyncio import AbstractEventLoop
from typing import List, Optional, Iterable
from urllib.parse import quote

from aiohttp import ClientSession, ClientResponse

from acapelladb.utils.errors import CasError, TransactionNotFoundError, TransactionCompletedError, KvError, \
    AuthenticationFailedError


def key_to_str(key: Iterable[str]) -> str:
    return ':'.join(quote(part) for part in key)


def entry_url(api_prefix: str, partition: List[str], clustering: Optional[List[str]] = None) -> str:
    if clustering is None or len(clustering) == 0:
        return f'{api_prefix}/v2/kv/keys/{key_to_str(partition)}'
    return f'{api_prefix}/v2/kv/partition/{key_to_str(partition)}/clustering/{key_to_str(clustering)}'


def raise_if_error(code: int):
    if code == 200:
        return
    if code == 401:
        raise AuthenticationFailedError()
    if code == 408:
        raise TimeoutError()
    if code == 409:
        raise CasError()
    if code == 410:
        raise TransactionNotFoundError()
    if code == 412:
        raise TransactionCompletedError()
    raise KvError(f'Unexpected server error with code {code}')


class AsyncSession(object):
    def __init__(self, session: ClientSession = None, loop: AbstractEventLoop = None, base_url: str = ''):
        self._session = session or ClientSession()
        self._loop = loop or asyncio.get_event_loop()
        self._base_url = base_url
        self._auth = None

    async def _request(self, method, url, **kwargs):
        kwargs = {'auth': self._auth, **kwargs}
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return await self._session.request(method, self._base_url + url, **kwargs)

    async def get(self, url, **kwargs) -> ClientResponse:
        return await self._request('get', url, **kwargs)

    async def options(self, url, **kwargs) -> ClientResponse:
        return await self._request('options', url, **kwargs)

    async def head(self, url, **kwargs) -> ClientResponse:
        return await self._request('head', url, **kwargs)

    async def post(self, url, data=None, json=None, **kwargs) -> ClientResponse:
        return await self._request('post', url, data=data, json=json, **kwargs)

    async def put(self, url, data=None, **kwargs) -> ClientResponse:
        return await self._request('put', url, data=data, **kwargs)

    async def patch(self, url, data=None, **kwargs) -> ClientResponse:
        return await self._request('patch', url, data=data, **kwargs)

    async def delete(self, url, **kwargs) -> ClientResponse:
        return await self._request('delete', url, **kwargs)

    def set_cookie(self, cookies):
        self._session.cookie_jar.update_cookies(cookies)

    def set_auth(self, auth):
        self._auth = auth



from acapelladb.Cursor import Cursor
from acapelladb.Entry import Entry
from acapelladb.Session import Session
from acapelladb.Transaction import Transaction
from acapelladb.TransactionContext import TransactionContext

name = "acapelladb"



from typing import List, Optional

from acapelladb.Cursor import Cursor
from acapelladb.Transaction import Transaction
from acapelladb.utils.assertion import check_key, check_nrw
from acapelladb.utils.collections import remove_none_values
from acapelladb.utils.http import AsyncSession, key_to_str, raise_if_error


class Tree(object):
    def __init__(self, session: AsyncSession, name: List[str], n: int, r: int, w: int, api_prefix: str):
        check_key(name)
        check_nrw(n, r, w)
        self._name = name
        self._session = session
        self._n = n
        self._r = r
        self._w = w
        self._api_prefix = api_prefix

    async def get_cursor(self, key: List[str], transaction: Optional[Transaction] = None) -> Cursor:
        """
        Получение значения по указанному ключу в дереве.
        :param key: ключ
        :param transaction если указан, курсор привязывается к транзакции
        :return: Cursor для указанного ключа
        """
        cursor = self.cursor(key, transaction)
        await cursor.get()
        return cursor

    def cursor(self, key: List[str], transaction: Optional[Transaction] = None) -> Cursor:
        """
        Создание Cursor для указанного ключа в дереве. Не выполняет никаких запросов.
        Можно использовать, если нет необходимости знать текущие значение и версию.
        :param key: ключ
        :param transaction если указан, запрос выполняется в транзакции
        :return: Cursor для указанного ключа
        """
        tx_index = transaction.index if transaction is not None else None
        return Cursor(self._session, self._name, key, 0, None, None, self._n, self._r, self._w, tx_index,
                      self._api_prefix)

    async def range(self,
                    first: Optional[List[str]] = None,
                    last: Optional[List[str]] = None,
                    limit: int = 0,
                    transaction: Optional[Transaction] = None) -> List[Cursor]:
        """
        Возвращает отсортированный список ключей в дереве в указанный пределах.
        :param first: начальный ключ, не включается в ответ; по умолчанию - с первого
        :param last: последий ключ, включается в ответ; по умолчанию - до последнего включительно
        :param limit: максимальное количество ключей в ответе, начиная с первого; по умолчанию - нет ограничений
        :param transaction: если указан, запрос выполняется в транзакции
        :return: список объектов Cursor с данными 
        """
        first = first or []
        last = last or []
        tx_index = transaction.index if transaction is not None else None

        response = await self._session.get(
            f'{self._api_prefix}/v2/dt/{key_to_str(self._name)}/keys',
            params=remove_none_values({
                'from': key_to_str(first),
                'to': key_to_str(last),
                'limit': limit,
                'n': self._n,
                'r': self._r,
                'w': self._w,
                'transaction': tx_index,
            })
        )
        raise_if_error(response.status)

        body = await response.json()

        def cursor_init(k, v):
            return Cursor(self._session, self._name, k, 0, v, None, self._n, self._r, self._w, tx_index,
                          self._api_prefix)

        # TODO: версия в DT
        return [cursor_init(c['key'], c['value']) for c in body]



import base64
from typing import List, Union, Optional

from aiohttp import BasicAuth
from requests.adapters import DEFAULT_RETRIES
from urllib3 import Retry

from acapelladb.BatchManual import BatchManual
from acapelladb.Entry import Entry
from acapelladb.PartitionIndex import PartitionIndex
from acapelladb.Transaction import Transaction
from acapelladb.TransactionContext import TransactionContext
from acapelladb.Tree import Tree
from acapelladb.utils.assertion import check_key, check_clustering, check_limit
from acapelladb.utils.collections import remove_none_values
from acapelladb.utils.http import AsyncSession, raise_if_error, entry_url, key_to_str


class Session(object):
    def __init__(self, host: str = '127.0.0.1', port: int = 12000,
                 api_prefix: str = '/acapelladb'):
        """
        Создание HTTP-сессии для взаимодействия с KV. 
        
        :param host: хост
        :param port: порт
        """
        base_url = f'http://{host}:{port}'
        self._session = AsyncSession(base_url=base_url)
        self._api_prefix = api_prefix

    @property
    def api_prefix(self) -> str:
        return self._api_prefix

    @staticmethod
    def _format_basic_auth(user: str, password: str) -> str:
        user_and_password = user + ':' + password
        return base64.standard_b64encode(user_and_password.encode()).decode()

    async def login(self, user: Optional[str] = None, password: Optional[str] = None):
        """
        Вход под указанным пользователем.
        :param user: имя пользователя
        :param password: пароль пользователя
        """
        response = await self._session.post(
            '/auth/login',
            data={
                'enableCookies': 'true',
                'invalidateOld': 'false',
                'loginAndPassword': self._format_basic_auth(user, password)
            }
        )
        raise_if_error(response.status)
        body = await response.json()
        token = body['token']
        self._session.set_auth(BasicAuth(user, token))

    def transaction(self) -> TransactionContext:
        """
        Создание контекста транзакции для использования в блоке `async with`.
        При выходе из блока происходит автоматическое применение/откат транзакции, 
        в зависимости от наличия исключений. Возможно завершение транзакции вручную,
        тогда автоматическое завершение не произойдёт.
        Примеры использования:
        
        async with session.transaction() as tx:
            entry e = await tx.get(["some", "key"])
            await e.cas("new_value")
            // автоматически вызовется tx.commit()
        
        
        async with session.transaction() as tx:
            entry e = await tx.get(["some", "key"])
            raise RuntimeError() // автоматически вызовется tx.rollback()
            await e.cas("new_value")            
            
        :return: контекст транзакции
        """
        return TransactionContext(self._session, self._api_prefix)

    async def transaction_manual(self, index: Optional[int] = None) -> Transaction:
        """
        Создание объекта транзакции в "ручном режиме". Применение/откат транзакции лежит на клиентском коде.
        Следует использовать, только если не удаётся работать с транзакцией через `async with`.

        :param index: если указан, то транзакция не создаётся, а просто оборачивается в объект Transaction
        :return: созданная транзакция
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        if index is None:
            response = await self._session.post(f'{self.api_prefix}/v2/tx')
            raise_if_error(response.status)
            body = await response.json()
            index = int(body['index'])
        return Transaction(self._session, index)

    async def get_entry(self, partition: List[str], clustering: Optional[List[str]] = None,
                        n: int = 3, r: int = 2, w: int = 2) -> Entry:
        """
        Получение значения по указанному ключу вне транзакции.
        
        :param partition: распределительный ключ
        :param clustering: сортируемый ключ
        :param n: количество реплик
        :param r: количество ответов для подтверждения чтения
        :param w: количество ответов для подтверждения записи
        :return: Entry для указанного ключа с полученным значением
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        clustering = clustering or []
        entry = Entry(self._session, self._api_prefix, partition, clustering, 0, None, n, r, w, None)
        await entry.get()
        return entry

    async def get_version(self, partition: List[str], clustering: Optional[List[str]] = None,
                          n: int = 3, r: int = 2, w: int = 2) -> int:
        """
        Получение версии указанного ключа вне транзакции.

        :param partition: распределительный ключ
        :param clustering: сортируемый ключ
        :param n: количество реплик
        :param r: количество ответов для подтверждения чтения
        :param w: количество ответов для подтверждения записи
        :return: версия
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        clustering = clustering or []
        url = f'{entry_url(self._api_prefix, partition, clustering)}/version'
        response = await self._session.get(url, params={
            'n': n,
            'r': r,
            'w': w,
        })
        raise_if_error(response.status)
        body = await response.json()
        return int(body['version'])

    def entry(self, partition: List[str], clustering: Optional[List[str]] = None,
              n: int = 3, r: int = 2, w: int = 2) -> Entry:
        """
        Создание Entry для указанного ключа вне транзакции. Не выполняет никаких запросов.
        Можно использовать, если нет необходимости знать текущие значение и версию.
        
        :param partition: распределительный ключ
        :param clustering: сортируемый ключ
        :param n: количество реплик
        :param r: количество ответов для подтверждения чтения
        :param w: количество ответов для подтверждения записи
        :return: Entry для указанного ключа
        """
        clustering = clustering or []
        return Entry(self._session, self._api_prefix, partition, clustering, 0, None, n, r, w, None)

    async def range(self,
                    partition: List[str],
                    first: Optional[List[str]] = None,
                    last: Optional[List[str]] = None,
                    limit: Optional[int] = None,
                    prefix: Optional[List[str]] = None,
                    n: int = 3,
                    r: int = 2,
                    w: int = 2) -> List[Entry]:
        """
        Возвращает отсортированный список ключей в дереве в указанный пределах.
        :param partition: распределительный ключ
        :param first: начальный ключ, не включается в ответ; по умолчанию - с первого
        :param last: последий ключ, включается в ответ; по умолчанию - до последнего включительно
        :param limit: максимальное количество ключей в ответе, начиная с первого; по умолчанию - нет ограничений
        :param prefix: префикс, к которому должны принадлежать все ключи в выборке
        :param n: количество реплик
        :param r: количество ответов для подтверждения чтения
        :param w: количество ответов для подтверждения записи
        :return: список объектов Entry с данными
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        check_key(partition)
        check_clustering(first)
        check_clustering(last)
        check_limit(limit)
        check_clustering(prefix)

        url = f'{self.api_prefix}/v2/kv/partition/{key_to_str(partition)}'
        response = await self._session.get(url, params=remove_none_values({
            'from': first and key_to_str(first),
            'to': last and key_to_str(last),
            'limit': limit,
            'prefix': prefix and key_to_str(prefix),
            'n': n,
            'r': r,
            'w': w,
        }))
        raise_if_error(response.status)
        body = await response.json()
        return [Entry(self._session, self._api_prefix, partition, e['key'], e['version'], e.get('value'), n, r, w, None) for e in body]

    def tree(self, tree: List[str], n: int = 3, r: int = 2, w: int = 2) -> Tree:
        """
        Создание дерева DT.
        :param tree: имя дерева
        :param n: количество реплик
        :param r: количество ответов для подтверждения чтения
        :param w: количество ответов для подтверждения записи
        :return: Tree
        """
        return Tree(self._session, tree, n, r, w, self._api_prefix)

    def batch_manual(self) -> BatchManual:
        return BatchManual(self._session, self._api_prefix)

    def partition_index(self, partition: List[str]) -> PartitionIndex:
        return PartitionIndex(self._session, partition, self._api_prefix)



from typing import List, Optional


def check_key(key: List[str]):
    if len(key) == 0:
        raise ValueError("key must not be empty")
    for part in key:
        check_key_part(part)


def check_clustering(key: Optional[List[str]]):
    if key is None:
        return
    for part in key:
        check_key_part(part)


def check_key_part(part: str):
    if len(part) == 0:
        raise ValueError("All key parts must not be empty")


def check_nrw(n: int, r: int, w: int):
    if n <= 0:
        raise ValueError("N parameter must be greater than zero")
    if n > 127:
        raise ValueError("N parameter must be less than 128")
    if r <= 0:
        raise ValueError("R parameter must be greater than zero")
    if w <= 0:
        raise ValueError("W parameter must be greater than zero")
    if r > n:
        raise ValueError("R parameter must be less then or equal to N")
    if w > n:
        raise ValueError("W parameter must be less then or equal to N")


def check_limit(limit: Optional[int]):
    if limit is None:
        return
    if limit < 0:
        raise ValueError("limit parameter must be greater than or equal to zero")



from datetime import timedelta
from typing import List, Optional, Awaitable, Callable

import asyncio

from acapelladb.BatchBase import BatchBase
from acapelladb.utils.assertion import check_key, check_nrw
from acapelladb.utils.collections import remove_none_values
from acapelladb.utils.http import AsyncSession, raise_if_error, entry_url


class Entry(object):
    def __init__(self, session: AsyncSession, api_prefix: str, partition: List[str], clustering: List[str],
                 version: int, value: Optional[any], n: int, r: int, w: int, transaction: Optional[int]):
        """
        Создание объекта связанного с указанным ключом. Этот метод предназначен для внутреннего использования.
        """
        check_key(partition)
        check_nrw(n, r, w)
        self._session = session
        self._partition = partition
        self._clustering = clustering
        self._version = version
        self._value = value
        self._n = n
        self._r = r
        self._w = w
        self._transaction = transaction
        self._api_prefix = api_prefix

    async def get(self, watch: bool = False) -> Optional[any]:
        """
        Запрашивает текущее значение с сервера.
        Запоминает новые значение и версию.

        :param watch: если true, то блокирует ключ до конца транзакции
        :return: полученное значение
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, в которой выполняется операция, не найдена 
        :raise TransactionCompletedError: когда транзакция, в которой выполняется операция, уже завершена
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        url = entry_url(self._api_prefix, self._partition, self._clustering)
        response = await self._session.get(url, params=remove_none_values({
            'n': self._n,
            'r': self._r,
            'w': self._w,
            'transaction': self._transaction,
            'watch': str(watch),
        }))
        raise_if_error(response.status)
        body = await response.json()
        self._version = int(body['version'])
        self._value = body.get('value')
        return self._value

    async def listen(self, wait_version: Optional[int] = None, timeout: Optional[timedelta] = None) -> Optional[any]:
        """
        Ожидает, пока версия значения не превысит указанную. 
        Если ожидание завершилось успешно, запоминает новые значение и версию.
        
        :param wait_version: ожидаемая версия; если не указана, то используется текущая версия 
        :param timeout: время ожидания; если не указано, то используется значение по умолчанию
        :return: полученное значение
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, в которой выполняется операция, не найдена 
        :raise TransactionCompletedError: когда транзакция, в которой выполняется операция, уже завершена
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        if wait_version is None:
            wait_version = self._version
        timeout_seconds = int(timeout.total_seconds()) if timeout is not None else None

        url = entry_url(self._api_prefix, self._partition, self._clustering)
        response = await self._session.get(url, params=remove_none_values({
            'n': self._n,
            'r': self._r,
            'w': self._w,
            'transaction': self._transaction,
            'waitVersion': wait_version,
            'waitTimeout': timeout_seconds,
        }))
        raise_if_error(response.status)

        body = await response.json()
        self._version = int(body['version'])
        self._value = body.get('value')
        return self._value

    def set(self, new_value: Optional[any], reindex: bool = False,
            batch: Optional[BatchBase] = None) -> Awaitable[int]:
        """
        Устанавливает новое значение.
        Запоминает новые значение и версию.        
        
        :param new_value: новое значение
        :param reindex: переиндексировать ключ с новым значением?
        :param batch: батч, в котором нужно выполнить запрос
        :return: новая версия
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, в которой выполняется операция, не найдена 
        :raise TransactionCompletedError: когда транзакция, в которой выполняется операция, уже завершена
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        return single_or_batch(batch, self._set_single, self._set_batch, new_value, reindex)

    async def _set_batch(self, batch: BatchBase, new_value: Optional[any], reindex: bool):
        self._version = await batch.set(self.partition, self.clustering, new_value, reindex, self._n, self._r, self._w)
        self._value = new_value
        return self._version

    async def _set_single(self, new_value: Optional[any], reindex: bool):
        url = entry_url(self._api_prefix, self._partition, self._clustering)
        response = await self._session.put(url, params=remove_none_values({
            'n': self._n,
            'r': self._r,
            'w': self._w,
            'transaction': self._transaction,
            'reindex': str(reindex),
        }), json=new_value)
        raise_if_error(response.status)
        body = await response.json()
        self._version = int(body['version'])
        self._value = new_value
        return self._version

    def cas(self, new_value: Optional[any], old_version: Optional[int] = None, reindex: bool = False,
            batch: Optional[BatchBase] = None) -> Awaitable[int]:
        """
        Устанавливает новое значение при совпадении версий.
        
        :param new_value: новое значение
        :param old_version: старая версия; если не указана, то используется текущая версия
        :param reindex: переиндексировать ключ с новым значением?
        :param batch: батч, в котором нужно выполнить запрос
        :return: новая версия
        :raise TimeoutError: когда время ожидания запроса истекло
        :raise TransactionNotFoundError: когда транзакция, в которой выполняется операция, не найдена 
        :raise TransactionCompletedError: когда транзакция, в которой выполняется операция, уже завершена
        :raise CasError: когда текущая версия значения не совпала с указанной
        :raise KvError: когда произошла неизвестная ошибка на сервере
        """
        if old_version is None:
            old_version = self._version
        return single_or_batch(batch, self._cas_single, self._cas_batch, new_value, old_version, reindex)

    async def _cas_batch(self, batch: BatchBase, new_value: Optional[any], old_version: int, reindex: bool):
        self._version = await batch.cas(
            self.partition, self.clustering, new_value, old_version, reindex, self._n, self._r, self._w
        )
        self._value = new_value
        return self._version

    async def _cas_single(self, new_value: Optional[any], old_version: int, reindex: bool):
        url = entry_url(self._api_prefix, self._partition, self._clustering)
        response = await self._session.put(url, params=remove_none_values({
            'n': self._n,
            'r': self._r,
            'w': self._w,
            'transaction': self._transaction,
            'oldVersion': old_version,
            'reindex': str(reindex),
        }), json=new_value)
        raise_if_error(response.status)
        body = await response.json()
        self._version = int(body['version'])
        self._value = new_value
        return self._version

    @property
    def value(self) -> Optional[any]:
        """
        :return: значение 
        """
        return self._value

    @property
    def version(self) -> int:
        """
        :return: версия 
        """
        return self._version

    @property
    def partition(self) -> List[str]:
        """
        :return: распределительный ключ
        """
        return self._partition

    @property
    def clustering(self) -> List[str]:
        """
        :return: сортируемый ключ
        """
        return self._clustering


def single_or_batch(batch: Optional[BatchBase], fn_single: Callable, fn_batch: Callable, *args) -> Awaitable:
    if batch is None:
        return fn_single(*args)
    else:
        return asyncio.ensure_future(fn_batch(batch, *args))



from typing import TypeVar, Dict, Optional

K = TypeVar('K')
V = TypeVar('V')


def remove_none_values(c: Dict[K, Optional[V]]) -> Dict[K, V]:
    return {k: v for k, v in c.items() if v is not None}



#!/usr/bin/env python

from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

setup(
    name='acapelladb',
    version='0.3.8',
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='Python client for AcapellaDB database',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
    url='https://github.com/AcapellaSoft/AcapellaDBClient-py',
    author='Acapella Soft',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'aiohttp == 3.4.4',
        'requests == 2.20.0',
    ],
)



from acapelladb.Transaction import Transaction

from acapelladb.utils.http import raise_if_error, AsyncSession


class TransactionContext(object):
    def __init__(self, session: AsyncSession, api_prefix: str):
        """        
        Создание контекста транзакции. Этот метод предназначен для внутреннего использования.
        """
        self._session = session
        self._api_prefix = api_prefix
        self._tx = None  # type: Transaction

    # для типизации
    def __enter__(self) -> Transaction:
        raise RuntimeError()

    async def __aenter__(self) -> Transaction:
        if self._tx is not None:
            raise RuntimeError("This transaction context already in entered state")

        response = await self._session.post(f'{self._api_prefix}/v2/tx')
        raise_if_error(response.status)
        body = await response.json()
        index = int(body['index'])
        tx = Transaction(self._session, index, self._api_prefix)
        self._tx = tx
        return tx

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            await self._tx.commit()
        else:
            await self._tx.rollback()
        self._tx = None



from enum import Enum


class IndexFieldType(Enum):
    string = 'string'
    number = 'number'


class IndexFieldOrder(Enum):
    ascending = 'ascending'
    descending = 'descending'


class IndexField(object):
    def __init__(self, name: str, type_: IndexFieldType, order: IndexFieldOrder):
        self.name = name
        self.type = type_
        self.order = order

    def to_json(self):
        return {
            'name': self.name,
            'type': self.type.value,
            'order': self.order.value,
        }

    @classmethod
    def from_json(cls, data) -> 'IndexField':
        return IndexField(
            name=data['name'],
            type_=IndexFieldType[data['type']],
            order=IndexFieldOrder[data['order']],
        )

    def __repr__(self) -> str:
        return f'IndexField(name={self.name}, type={self.type}, order={self.order})'

    def __eq__(self, other) -> bool:
        if not isinstance(other, IndexField):
            return False
        return self.name == other.name and self.type == other.type and self.order == other.order
